# Copyright 2022 Thomas Eboli (thomas.eboli12@gmail.com)
# 
# Permission is hereby granted, free of charge, to any person obtaining 
# a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, including without limitation 
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS IN THE SOFTWARE.


import numpy as np
cimport cython

# "cimport" is used to import special compile-time information
# about the numpy module (this is stored in a file numpy.pxd which is
# currently part of the Cython distribution).
cimport numpy as np

from cython.parallel cimport prange
from libc.math cimport fmin, fmax, fabs

# It's necessary to call "import_array" if you use any part of the
# numpy PyArray_* API. From Cython 3, accessing attributes like
# ".shape" on a typed Numpy array use this API. Therefore we recommend
# always calling "import_array" whenever you "cimport numpy"
np.import_array()

DTYPE = np.float32

# "ctypedef" assigns a corresponding compile-time type to DTYPE_t. For
# every type in the numpy module there's a corresponding compile-time
# type with a _t-suffix.
ctypedef np.float32_t DTYPE_t
ctypedef np.uint8_t uint8


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef DTYPE_t[:, ::1] transpose(DTYPE_t[:, ::1] X):
    """
    Simple transpose function for running the TI and FC 1D filters in both directions.

    Input:
        -- X, (M,N): The input array.

    Output:
        -- Xt, (N,M): The transposed array.
    """

    cdef Py_ssize_t M = X.shape[0]
    cdef Py_ssize_t N = X.shape[1]
    cdef Py_ssize_t i, j, k
    cdef DTYPE_t[:, ::1] Xt = np.empty((N, M), dtype=DTYPE)

    for i in prange(0, M, 1, nogil=True):
        for j in range(0, N, 1):
            Xt[j, i] = X[i, j]

    return Xt


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def chromatic_removal(np.ndarray[DTYPE_t, ndim=3] I_in, 
                      int L_hor, 
                      int L_ver,
                      np.ndarray[DTYPE_t, ndim=1] rho, 
                      DTYPE_t tau, 
                      DTYPE_t alpha_R, 
                      DTYPE_t alpha_B, 
                      DTYPE_t beta_R,
                      DTYPE_t beta_B, 
                      DTYPE_t gamma_1, 
                      DTYPE_t gamma_2):
    """
    Implements Alg. (5) in the paper, that embeds the three main stages:
        -- TI filtering
        -- FC filtering
        -- Arbitration

    Inputs:
        -- I_in, (M,N,3) array: The input image.
        -- L_hor, int:          The horizontal filter radius. 
        -- L_ver, int:          The vertical filter radius. 
        -- rho, (3) array:      The prefiltering coefficients in TI stage.
        -- tau, float:          The contrast threshold in FC stage.
        -- alpha_R, float:      The red chroma weight in FC stage.
        -- alpha_B, float:      The blue chroma weight in FC stage.
        -- beta_R, float:       The red chroma gradient penalization in arbitration stage.
        -- beta_B, float:       The blue chroma gradient penalization in arbitration stage.
        -- gamma_1, float:      The first contrast function parameter in arbitration stage.
        -- gamma_2, float:      The second contrast function parameter in arbitration stage.

    Outputs:
        -- I_out, (M,N,3) array: The final arbitrated image.
        -- I_TI, (M,N,3) array:  The TI intermediate image.
        -- I_FC, (M,N,3) array:  The FC intermediate image.
    """


    ##########################
    ##### Initialization #####
    ##########################

    cdef Py_ssize_t M, N, i, j, k
    M = I_in.shape[0]
    N = I_in.shape[1]

    ## Input arrays
    cdef DTYPE_t[:, ::1] R_in = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] G_in = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] B_in = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] Y_in = np.empty((M, N), dtype=DTYPE)

    ## TI and FC intermediate arrays
    ### Horizontal passes
    cdef DTYPE_t[:, ::1] K_h_rFC   = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] K_h_rTI = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] R_h_max = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] R_h_min = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] K_h_bFC   = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] K_h_bTI = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] B_h_max = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] B_h_min = np.empty((M, N), dtype=DTYPE)

    ### Vertical passes (TRANSPOSED ARRAYS!)
    cdef DTYPE_t[:, ::1] K_v_rFC   = np.empty((N, M), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] K_v_rTI = np.empty((N, M), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] R_v_max = np.empty((N, M), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] R_v_min = np.empty((N, M), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] K_v_bFC   = np.empty((N, M), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] K_v_bTI = np.empty((N, M), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] B_v_max = np.empty((N, M), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] B_v_min = np.empty((N, M), dtype=DTYPE)

    ## Output arrays
    cdef DTYPE_t[:, :, ::1] I_TI = np.empty((M, N, 3), dtype=DTYPE)
    cdef DTYPE_t[:, :, ::1] I_FC = np.empty((M, N, 3), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] K_rout = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] K_bout = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, :, ::1] I_out = np.empty((M, N, 3), dtype=DTYPE)
 
    ### Init YUV image and extracts input channels
    for k in prange(0, M*N, 1, nogil=True):
        j = k % N
        i = (k-j) / N
        R_in[i, j] = I_in[i, j, 0]
        G_in[i, j] = I_in[i, j, 1]
        B_in[i, j] = I_in[i, j, 2]
        Y_in[i, j] = 0.299 * R_in[i, j] + 0.587 * G_in[i, j] + 0.114 * B_in[i, j]



    ###############################
    ##### TI and FC Filtering #####
    ###############################

    ## Horizontal pass (Alg. (1) and (2))
    ti_and_fc_filtering1D(R_in, G_in, Y_in, L_hor, rho, alpha_R, tau, K_h_rFC, K_h_rTI, R_h_max, R_h_min)
    ti_and_fc_filtering1D(B_in, G_in, Y_in, L_hor, rho, alpha_B, tau, K_h_bFC, K_h_bTI, B_h_max, B_h_min)

    # Vertical pass (Alg. (1) and (2) with transposed arrays)
    ti_and_fc_filtering1D(transpose(R_in), transpose(G_in), transpose(Y_in), L_ver, rho, alpha_R, tau, K_v_rFC, K_v_rTI, R_v_max, R_v_min)
    ti_and_fc_filtering1D(transpose(B_in), transpose(G_in), transpose(Y_in), L_ver, rho, alpha_B, tau, K_v_bFC, K_v_bTI, B_v_max, B_v_min)

    # Invert transposition of the vertical pass
    K_v_rFC = transpose(K_v_rFC)
    K_v_bFC = transpose(K_v_bFC)
    K_v_rTI = transpose(K_v_rTI)
    K_v_bTI = transpose(K_v_bTI)
    R_v_max = transpose(R_v_max)
    B_v_max = transpose(B_v_max)
    R_v_min = transpose(R_v_min)
    B_v_min = transpose(B_v_min)



    #############################
    ####### Image merge #########
    #############################

    # Build the 2D images from the vertically and the horizontally FC filtered images (Eqs. (12))
    # Kb
    cdef DTYPE_t[:, ::1] K_bFC = K_v_bFC
    for k in prange(0, M*N, 1, nogil=True):
        j = k % N
        i = (k-j) / N
        if fabs(K_h_bFC[i, j]) < fabs(K_v_bFC[i, j]):
            K_bFC[i, j] = K_h_bFC[i, j]

    #Kr
    cdef DTYPE_t[:, ::1] K_rFC = K_v_rFC
    for k in prange(0, M*N, 1, nogil=True):
        j = k % N
        i = (k-j) / N
        if fabs(K_h_rFC[i, j]) < fabs(K_v_rFC[i, j]):
            K_rFC[i, j] = K_h_rFC[i, j]

    # Build the 2D images from the vertically and the horizontally TI filtered images (Eqs. (13))
    # Kb
    cdef DTYPE_t[:, ::1] K_bTI = K_v_bTI
    cdef DTYPE_t[:, ::1] B_max = B_v_max
    cdef DTYPE_t[:, ::1] B_min = B_v_min
    for k in prange(0, M*N, 1, nogil=True):
        j = k % N
        i = (k-j) / N
        if fabs(K_h_bTI[i, j]) < fabs(K_v_bTI[i, j]):
            K_bTI[i, j] = K_h_bTI[i, j]
            B_max[i, j] = B_h_max[i, j]
            B_min[i, j] = B_h_min[i, j]

    # Kr
    cdef DTYPE_t[:, ::1] K_rTI = K_v_rTI
    cdef DTYPE_t[:, ::1] R_max = R_v_max
    cdef DTYPE_t[:, ::1] R_min = R_v_min
    for k in prange(0, M*N, 1, nogil=True):
        j = k % N
        i = (k-j) / N
        if fabs(K_h_rTI[i, j]) < fabs(K_v_rTI[i, j]):
            K_rTI[i, j] = K_h_rTI[i, j]
            R_max[i, j] = R_h_max[i, j]
            R_min[i, j] = R_h_min[i, j]



    #############################
    ####### Arbitration #########
    #############################

    ## Contrast arbitration (Alg. 4)
    arbitration(R_in, G_in, K_rTI, K_rFC, R_max, R_min, beta_R, gamma_1, gamma_2, L_hor, L_ver, K_rout)
    arbitration(B_in, G_in, K_bFC, K_bTI, B_max, B_min, beta_B, gamma_1, gamma_2, L_hor, L_ver, K_bout)


    #############################
    ##### Color conversion ######
    #############################

    ## Luma-chroma to RGB conversion (Eq. (19))
    for k in prange(0, M*N, 1, nogil=True):
        j = k % N
        i = (k - j) / N
        # Transient improvement image
        I_TI[i, j, 0] = K_rTI[i, j] + G_in[i, j]
        I_TI[i, j, 1] = G_in[i, j]
        I_TI[i, j, 2] = K_bTI[i, j] + G_in[i, j]
        # False color image
        I_FC[i, j, 0] = K_rFC[i, j] + G_in[i, j]
        I_FC[i, j, 1] = G_in[i, j]
        I_FC[i, j, 2] = K_bFC[i, j] + G_in[i, j]
        # Arbitrated image
        I_out[i, j, 0] = K_rout[i, j] + G_in[i, j]
        I_out[i, j, 1] = G_in[i, j]
        I_out[i, j, 2] = K_bout[i, j] + G_in[i, j]

    return np.clip(np.asarray(I_out), 0, 1), np.clip(np.asarray(I_TI), 0, 1), np.clip(np.asarray(I_FC), 0, 1)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef void grad(DTYPE_t[:, ::1] X, 
               Py_ssize_t M, 
               Py_ssize_t N, 
               DTYPE_t[:, ::1] grad_X) nogil:
    """
    Simple finite-difference filter to implement the contrast function in Eqs. (8) and (10) 
    """
    cdef Py_ssize_t i, j, k
    for k in prange(0, M*N, 1, nogil=True):
        j = k % N
        i = (k - j) / N 
        grad_X[i, j] = X[i, j] - X[i, j - 1]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef int sign(DTYPE_t input) nogil:
    """
    Sign function for the binary weight computation in Eq. (9)
    """
    if input > 0:
        return 1
    elif input < 0:
        return -1
    else:
        return 0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef DTYPE_t clip(DTYPE_t K, 
                  DTYPE_t K_ref) nogil:
    """
    Clipping of the TI slice in Eq. (7)
    """
    if K_ref > 0:
        return fmin(K, K_ref)
    elif K_ref < 0:
        return fmax(K, K_ref)
    else:
        return K_ref


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void ti_and_fc_filtering1D(DTYPE_t[:, ::1] X_in, 
                                DTYPE_t[:, ::1] G_in, 
                                DTYPE_t[:, ::1] Y_in, 
                                int L,
                                DTYPE_t[::1] rho, 
                                DTYPE_t alpha_X, 
                                DTYPE_t tau, 
                                DTYPE_t[:, ::1] K_h_FC,
                                DTYPE_t[:, ::1] K_h_TI, 
                                DTYPE_t[:, ::1] X_h_max, 
                                DTYPE_t[:, ::1] X_h_min):
    """
    1D sucessive computation of the TI and FC filtered images in the horizontal direction for a single
    color channel among red and blue. It implements Algs. (1) and (2) that are fused together in this
    function for efficiency's sake.

    Inputs:
        -- X_in, (M,N) array: The input red or blue color.
        -- G_in, (M,N) array: The input green color.
        -- Y_in, (M,N) array: The input luma channel in a YUV decomposition of the input.
        -- L, int:            The filter radius.
        -- rho, (3) array:    The prefiltering coefficients in TI stage.
        -- tau, float:        The contrast threshold in FC stage.
        -- alpha_R, float:    The red chroma weight in FC stage.
        -- alpha_B, float:    The blue chroma weight in FC stage.

    Outputs:
        -- K_h_TI, (M,N) array:  The TI horizontally-filtered intermediate chroma channel.
        -- K_h_FC, (M,N) array:  The FC horizontally-filtered intermediate chroma channel.
        -- X_h_max, (M,N array:  The horizontal maxima used later for arbitration.
        -- X_h_min, (M,N) array: The horizontal minima used later for arbitration.
    """

    ##########################
    ##### Initialization #####
    ##########################


    cdef Py_ssize_t M, N, i, j
    cdef int l = 0
    cdef int LL = 2 * L + 1
    cdef DTYPE_t eps = 1e-8

    M = X_in.shape[0]
    N = X_in.shape[1]

    ## All the local variables that are scalars or 2L+1 subarrays in the IPOL publications have
    ## here an additional dimension of size M (the number of rows): This is for parallelizing the method
    ## to make it significantly much faster in practice.
    cdef DTYPE_t[:, ::1] hat_X_h_TImax = np.empty((M, LL), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] hat_X_h_TImin = np.empty((M, LL), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] hat_X_h_TI = np.empty((M, LL), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] hat_X_h_pf = np.empty((M, LL), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] hat_K_h_TI = np.empty((M, LL), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] W_K = np.empty((M, LL), dtype=DTYPE)
    cdef DTYPE_t[::1] W_K_sum = np.empty(M, dtype=DTYPE)
    cdef DTYPE_t[::1] X_Emax = np.empty(M, dtype=DTYPE)
    cdef DTYPE_t[::1] X_Emin = np.empty(M, dtype=DTYPE)
    cdef DTYPE_t[::1] X_Wmax = np.empty(M, dtype=DTYPE)
    cdef DTYPE_t[::1] X_Wmin = np.empty(M, dtype=DTYPE)

    cdef DTYPE_t[:, ::1] grad_X = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] grad_G = np.empty((M, N), dtype=DTYPE)
    grad(X_in, M, N, grad_X)
    grad(G_in, M, N, grad_G)


    ##############################################
    ## This loop is partially paralellized to be
    ## memory-efficient since it involves many
    ## large 3D tensors, especially for images of 
    ## more than 3Mpixels for instance. Yet
    ## it may be made fully paralel if needs be.
    ##############################################

    for i in prange(L, M - L, 1, nogil=True):
        for j in range(L, N - L, 1):
            W_K_sum[i] = 0
            X_Emax[i] = X_in[i, j]
            X_Emin[i] = X_in[i, j]
            X_Wmax[i] = X_in[i, j]
            X_Wmin[i] = X_in[i, j]


            ########################################
            ##### Initial max/min computations #####
            ########################################

            ## Compute min and max on windows (Eqs. (2a) to (2d))
            for l in range(0, L, 1):
                if X_in[i, j + l + 1] > X_Emax[i]:
                    X_Emax[i] = X_in[i, j + l + 1]
                if X_in[i, j + l + 1] < X_Emin[i]:
                    X_Emin[i] = X_in[i, j + l + 1]
                if X_in[i, j - L + l] > X_Wmax[i]:
                    X_Wmax[i] = X_in[i, j - L + l]
                if X_in[i, j - L + l] < X_Wmin[i]:
                    X_Wmin[i] = X_in[i, j - L + l]

            ## Select the best couple of values (Eq. (2e))
            if (X_Emax[i] - X_Wmin[i]) >= (X_Wmax[i] - X_Emin[i]):
                X_h_max[i, j] = X_Emax[i]
                X_h_min[i, j] = X_Wmin[i]
            else:
                X_h_max[i, j] = X_Wmax[i]
                X_h_min[i, j] = X_Emin[i]

            ## Iteration over the slices' pixels: actual filtering.
            for l in range(0, LL, 1):

                ########################
                ##### TI Filtering #####
                ########################

                ## Main TI process (Eq. (1)) and Eq. (3))
                if X_in[i, j] > G_in[i, j]:
                    hat_X_h_pf[i, l] = rho[0] * X_h_max[i, j] + rho[1] * X_in[i, j + l - L] + rho[2] * X_h_min[i, j]
                    hat_X_h_TImax[i, l] = X_in[i, j + l - L]
                    hat_X_h_TImin[i, l] = fmax(X_h_min[i, j], G_in[i, j + l - L])
                else:
                    hat_X_h_pf[i, l] = rho[0] * X_h_min[i, j] + rho[1] * X_in[i, j + l - L] + rho[2] * X_h_max[i, j]
                    hat_X_h_TImax[i, l] = fmin(X_h_max[i, j], G_in[i, j + l - L])
                    hat_X_h_TImin[i, l] = X_in[i, j + l - L]

                ## Clipping the filtered imaged in the admissible set on values (Eq. (4))
                if hat_X_h_pf[i, l] > hat_X_h_TImax[i, l]:
                    hat_X_h_TI[i, l] = hat_X_h_TImax[i, l]
                elif hat_X_h_pf[i, l] < hat_X_h_TImin[i, l]:
                    hat_X_h_TI[i, l] = hat_X_h_TImin[i, l]
                else:
                    hat_X_h_TI[i, l] = hat_X_h_pf[i, l]

                ########################
                ##### FC Filtering #####
                ########################

                # RGB to luma-chroma conversion (Eq. (6))
                hat_K_h_TI[i, l] = hat_X_h_TI[i, l] - G_in[i, j + l - L]   # (2L+1)

                ##### False color filtering
                ## Chromaticity sign (Eq. (10))
                if (sign(hat_K_h_TI[i, L]) == sign(hat_K_h_TI[i, l])) or (fabs(hat_K_h_TI[i, l]) < tau):
                    W_K[i, l] = 1
                else:
                    W_K[i, l] = 0

                ## Gradients' weights (Eqs. (9) and (11))
                W_K[i, l] /= fabs(grad_G[i, j + l - L]) + \
                             fmax(fabs(grad_X[i, j + l - L]), alpha_X * fabs(hat_K_h_TI[i, l])) + \
                             fabs(Y_in[i, j] - Y_in[i, j + l - L])  + eps

                ## Compute numerator and denominator with clipping (Eqs. (7) and (10))
                K_h_FC[i, j] += W_K[i, l] * clip(hat_K_h_TI[i, l], hat_K_h_TI[i, L])
                W_K_sum[i] += W_K[i, l]

            ## Linear filtering (Eqs. (7))
            K_h_FC[i, j] /= W_K_sum[i] + eps  ## adding eps for numerical stability, just in case.

            ## Save K_TI for arbitration (Eq. (5))
            K_h_TI[i, j] = hat_K_h_TI[i, L]



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void contrast1D(DTYPE_t[:, ::1] X_in, 
                     DTYPE_t[:, ::1] G_in, 
                     DTYPE_t beta_X, 
                     int L, 
                     DTYPE_t[:, ::1] X_h_contrast):
    """
    Implements Alg. (3) in the paper that computes the directional contrasts.

    Inputs:
        -- X_in, (M, N) array:      The input color channel.
        -- G_in, (M, N) array:      The input green channel.
        -- beta_X, float:           The color gradiennt penalization.
        -- L, int:                  The horizontal filter radius.

    Outputs:
        -- X_h_contrast, (M, M) array: The local contrasts in the horizontal direction.
    """

    cdef Py_ssize_t M, N, i, j, k
    cdef int l = 0
    M = X_in.shape[0]
    N = X_in.shape[1]

    cdef DTYPE_t[:, ::1] X_Emax = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] X_Emin = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] X_Wmax = np.empty((M, N), dtype=DTYPE)
    cdef DTYPE_t[:, ::1] X_Wmin = np.empty((M, N), dtype=DTYPE)

    ## Compute the local max and min in the horizontal pass
    for k in prange(0, (M - 2*L)*(N - 2*L), 1, nogil=True):
        j = k % (N - 2*L)
        i = (k - j) / (N - 2*L)
        i = i + L  # i in [L,N-L]
        j = j + L  # j in [L,N-L]
        X_Emax[i, j] = X_in[i, j]
        X_Emin[i, j] = X_in[i, j]
        X_Wmax[i, j] = X_in[i, j]
        X_Wmin[i, j] = X_in[i, j]

        ## Recompute updated local max and min (Eqs. (17a) to (17d))
        for l in range(0, L, 1):
            if (X_in[i, j + l] - beta_X * fabs(X_in[i, j + l] - G_in[i, j + l])) > X_Emax[i,j]:
                X_Emax[i, j] = X_in[i, j + l] - beta_X * fabs(X_in[i, j + l] - G_in[i, j + l])

            if (X_in[i, j + l] + beta_X * fabs(X_in[i, j + l] - G_in[i, L + l])) < X_Emin[i,j]:
                X_Emin[i, j] = X_in[i, j + l] + beta_X * fabs(X_in[i, j + l] - G_in[i, j + l])

            if (X_in[i, j - L + l + 1] - beta_X * fabs(X_in[i, j - L + l + 1] - G_in[i, j - L + l + 1])) > X_Wmax[i,j]:
                X_Wmax[i, j] = X_in[i, j - L + l + 1] - beta_X * fabs(X_in[i, j - L + l + 1] - G_in[i, j - L + l + 1])

            if (X_in[i, j - L + l + 1] + beta_X * fabs(X_in[i, j - L + l + 1] - G_in[i, j - L + l + 1])) < X_Wmin[i,j]:
                X_Wmin[i, j] = X_in[i, j - L + l + 1] + beta_X * fabs(X_in[i, j - L + l + 1] - G_in[i, j - L + l + 1])

        ## Select the best couple of values (Eq. (17e))
        if (X_Emax[i, j] - X_Wmin[i, j]) >= (X_Wmax[i, j] - X_Emin[i, j]):
            X_h_contrast[i, j] = X_Emax[i, j] - X_Wmin[i, j]
        else:
            X_h_contrast[i, j] = X_Wmax[i, j] - X_Emin[i, j]



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void arbitration(DTYPE_t[:, ::1] X_in, 
                      DTYPE_t[:, ::1] G_in, 
                      DTYPE_t[:, ::1] K_TI, 
                      DTYPE_t[:, ::1] K_FC, 
                      DTYPE_t[:, ::1] X_max,
                      DTYPE_t[:, ::1] X_min, 
                      DTYPE_t beta_X, 
                      DTYPE_t gamma_1,
                      DTYPE_t gamma_2, 
                      int L_hor, 
                      int L_ver, 
                      DTYPE_t[:, ::1] K_out):
    """
    Implements Alg. (4) in the paper, that fuses of the TI and FC merged images.

    Inputs:
        -- X_in, (M,N) array:  The input corresponding color channel.
        -- G_in, (M,N) array:  The input green channel.
        -- K_TI, (M,N) array:  The TI filtered chroma channel.
        -- K_FC, (M,N) array:  The FC filtered chroma channel.
        -- X_max, (M,N) array: The local maxima computed in the TI stage and further merged.
        -- X_min, (M,N) array: The local minima computed in the TI stage and further merged.
        -- beta_X, float:      The chroma gradient penalization in arbitration stage.
        -- gamma_1, float:     The first contrast function parameter in arbitration stage.
        -- gamma_2, float:     The second contrast function parameter in arbitration stage.
        -- L_hor, int:         The horizontal filter radius. 
        -- L_ver, int:         The vertical filter radius. 

    Output:
        -- K_out, (M,N,3) array: The final arbitrated chroma channel.
    """

    ##########################
    ##### Initialization #####
    ##########################

    cdef Py_ssize_t M, N, i, j, k
    cdef int l = 0
    M = X_in.shape[0]
    N = X_in.shape[1]

    cdef DTYPE_t[:, ::1] X_h_contrast = np.empty((M, N), dtype=DTYPE)  # (M, N), horizontal pass
    cdef DTYPE_t[:, ::1] X_v_contrast = np.empty((N, M), dtype=DTYPE)  # (N, M), vertical pass TRANSPOSED!
    cdef DTYPE_t [::1] alpha_K = np.empty(M, dtype=DTYPE)


    ######################################
    ##### Compute contrast (Alg. 3) ######
    ######################################

    ## Compute the local max and min in the horizontal pass (Eqs. (17a) to (17e))
    contrast1D(X_in, G_in, beta_X, L_hor, X_h_contrast)
    contrast1D(transpose(X_in), transpose(G_in), beta_X, L_ver, X_v_contrast)
    X_v_contrast = transpose(X_v_contrast)

    ## Compute X_contrast  from the two directional estimates (Eq. (18))
    cdef DTYPE_t[:, ::1] X_contrast = X_v_contrast
    for k in prange(0, M*N, 1, nogil=True):
        j = k % N
        i = (k-j) / N
        if X_h_contrast[i, j] > X_v_contrast[i, j]:
            X_contrast[i, j] = X_h_contrast[i, j]


    #############################
    ######## Convex sum #########
    #############################

    ## Convex summation
    for k in prange(0, M*N, 1, nogil=True):
        j = k % N
        i = (k-j) / N
        ## Compute the arbitration weight (Eq. (16)) and the final filtered image (Eq. (15))
        alpha_K[i] = fmin(fmax(X_contrast[i, j], 0.0) / fmin(fmax(X_max[i, j] - X_min[i, j], gamma_2), gamma_1), 1.0)
        K_out[i, j] = (1 - alpha_K[i]) * K_TI[i, j] + alpha_K[i] * K_FC[i, j]
