# Copyright 2022 Thomas Eboli (thomas.eboli12@gmail.com)
# 
# Permission is hereby granted, free of charge, to any person obtaining 
# a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, including without limitation 
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS IN THE SOFTWARE.


import os
import time
from argparse import ArgumentParser

import numpy as np
from skimage import io, img_as_float32, img_as_ubyte
import cv2

import filter_cython


## Parameters
parser = ArgumentParser()

# Main parameters
parser.add_argument('--input', type=str)
parser.add_argument('--output', type=str)

# Advanced parameters
parser.add_argument('--rho_0', default=-0.25, type=float)
parser.add_argument('--rho_1', default=1.375, type=float)
parser.add_argument('--rho_2', default=-0.125, type=float)
parser.add_argument('--L_hor', default=7, type=int)
parser.add_argument('--L_ver', default=4, type=int)
parser.add_argument('--alpha_R', default=0.5, type=float)
parser.add_argument('--alpha_B', default=1.0, type=float)
parser.add_argument('--beta_R', default=1.0, type=float)
parser.add_argument('--beta_B', default=0.25, type=float)
parser.add_argument('--tau', default=15./256, type=float)
parser.add_argument('--gamma_1', default=128./256, type=float)
parser.add_argument('--gamma_2', default=64./256, type=float)


args = parser.parse_args()

img = img_as_float32(io.imread(args.input))

# read parameters
rho = np.array([args.rho_0, args.rho_1, args.rho_2], dtype=img.dtype)
L_hor = args.L_hor
L_ver = args.L_ver
alpha_R = args.alpha_R
alpha_B = args.alpha_B
beta_R = args.beta_R
beta_B = args.beta_B
tau = args.tau
gamma_1 = args.gamma_1
gamma_2 = args.gamma_2


print('Use the parameters:')
print('  rho:    ', list(rho))
print('  L_ver:  ', L_ver)
print('  L_hor:  ', L_hor)
print('  alpha_R:', alpha_R)
print('  alpha_B:', alpha_B)
print('  beta_R: ', beta_R)
print('  beta_B: ', beta_B)
print('  gamma_1:', gamma_1)
print('  gamma_2:', gamma_2)
print('  tau:', tau)


img = np.pad(img, [(L_ver, L_ver), (L_hor, L_hor), (0,0)], mode='edge')

print('Start restoration...')
tic = time.time()
impred, im_TI, im_FC = filter_cython.chromatic_removal(img, L_hor, L_ver, rho, 
        tau, alpha_R, alpha_B, beta_R, beta_B, gamma_1, gamma_2)
toc = time.time()
print("Elapsed time: %2.2f sec" % (toc - tic))

impred = impred[L_ver:-L_ver, L_hor:-L_hor, :]
im_TI = im_TI[L_ver:-L_ver, L_hor:-L_hor, :]
im_FC = im_FC[L_ver:-L_ver, L_hor:-L_hor, :]

im_TI = img_as_ubyte(im_TI)
im_FC = img_as_ubyte(im_FC)
impred = img_as_ubyte(impred)

im_TI = cv2.cvtColor(im_TI, cv2.COLOR_RGB2BGR)
im_FC = cv2.cvtColor(im_FC, cv2.COLOR_RGB2BGR)
impred = cv2.cvtColor(impred, cv2.COLOR_RGB2BGR)

suffix = args.output[-4:]
cv2.imwrite(args.output[:-4] + '_TI' + suffix, im_TI)
cv2.imwrite(args.output[:-4] + '_FC' + suffix, im_FC)
cv2.imwrite(args.output, impred)
# io.imsave(args.output[:-4] + '_TI' + suffix, img_as_ubyte(im_TI))
# io.imsave(args.output[:-4] + '_FC' + suffix, img_as_ubyte(im_FC))
# io.imsave(args.output, img_as_ubyte(impred))

print('Done')
